/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.core.instancer;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.UUID;

import org.mozilla.javascript.Context;
import org.mozilla.javascript.NativeJavaObject;
import org.mozilla.javascript.NativeObject;
import org.mozilla.javascript.ScriptRuntime;
import org.mozilla.javascript.Scriptable;
import org.mozilla.javascript.WrapFactory;

import com.unitesk.requality.core.Attribute;
import com.unitesk.requality.core.AttributeType;
import com.unitesk.requality.core.ITreeStorage;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.AbstractTreeStorage;
import com.unitesk.requality.core.TreeNode;

public class TreeInstancer extends AbstractTreeStorage {
    private ITreeStorage storage;
    private TreeDB srcTree;
    private Map<String, Object> initalAttribute = new HashMap<String, Object>();
    private Map<UUID, String> nodePredicates = new HashMap<UUID, String>();
    private Map<UUID, Collection<UUID>> toChildren = new LinkedHashMap<UUID, Collection<UUID>>();
    private Map<UUID, Map<String,Object>> nodeAttribute = new HashMap<UUID, Map<String,Object>>();
    private Map<UUID, Map<String,Attribute>> changedAttribute = new HashMap<UUID, Map<String,Attribute>>();
    
    public TreeInstancer(TreeDB db, Map<String, Object> map) {
        this.initalAttribute = map;
        this.storage = db.getStorage();
        this.srcTree = db;
    }

    @Override
    public UUID getRootNode() {
        if(!nodeAttribute.containsKey(storage.getRootNode())) {
            nodeAttribute.put(storage.getRootNode(), getAttributesMap(storage.getRootNode()));
        }
        if(!nodePredicates.containsKey(storage.getRootNode())) {
            nodePredicates.put(storage.getRootNode(), getPredicate(storage.getRootNode()));
        }
        return storage.getRootNode();
    }

    @Override
    public UUID getParent(UUID uuid) {
        return storage.getParent(uuid);
    }
    
    public String getErrors(){
    	UUID root = getRootNode();
    	Map<String,Object> attr = getAttributesMap(root);
    	if(attr!=null)
    	{
    		Object r = attr.get(TreeNode.ATTR_ERROR);
    		if(r != null)
    		return (String)r;
    	}
    	return "";
    }

    @Override
    public Collection<UUID> getChildren(UUID uuid) {
        if(toChildren.containsKey(uuid)) {
            return toChildren.get(uuid);
        }
        Collection<UUID> result = new ArrayList<UUID>(storage.getChildren(uuid));
        for(String qid : srcTree.getNode(uuid).getBaseNodes()) {
            if(qid.equals(srcTree.getNode(uuid).getQualifiedId())){
                result.add(uuid);
            }else{
                result.addAll(getChildren(srcTree.getNode(qid).getUUId()));
            }
        }
        
        Collection<UUID> tmp = new ArrayList<UUID>(result);
        for(UUID id: tmp) {
        	String s = checkPredicate(id);
            if(s.charAt(0)=='0') {
                result.remove(id);
            }
            s = s.substring(1);
            if(!s.equals(""))
            	registerError(id, "predicate error - "+s);
        }
        toChildren.put(uuid, result);
        return result;
    }
    
    private void addToErrorAttribute(UUID id, String value){
    	if(!changedAttribute.containsKey(id))
    		changedAttribute.put(id, new HashMap<String, Attribute>());
    	Map<String,Attribute> map = changedAttribute.get(id);
    	String val = "";
    	if(map.containsKey(TreeNode.ATTR_ERROR))
    		val = (String)map.get(TreeNode.ATTR_ERROR).getValue()+"\n";
    	map.put(TreeNode.ATTR_ERROR, new Attribute(AttributeType.STRING,TreeNode.ATTR_ERROR,val+value));
    	changedAttribute.put(id, map);

    	Map<String,Object> map2 = getAttributesMap(id);
    	String s = value;
    	if(map2.containsKey(TreeNode.ATTR_ERROR)){
    		s = (String) map2.get(TreeNode.ATTR_ERROR) + "\n" + s;
    	}
    	map2.put(TreeNode.ATTR_ERROR, s);
    	nodeAttribute.put(id, map2);
    }
    
    private void registerError(UUID id, String error){
    	UUID root = getRootNode();
    	String nid = getQualifyingId(id);
    	addToErrorAttribute(id, error);
    	addToErrorAttribute(root, "Node '" +nid+"' has errors: "+error);
    }

    private String getQualifyingId(UUID id) {
		String s = getNodeId(id);
		UUID par = getParent(id);
		if(par!=null){
			String qid = getQualifyingId(par);
			if(qid.equals("/"))
				qid = "";
			s= qid +"/"+s;
		}
		return s;
	}

	private String checkPredicate(UUID id) {
        String predicate = getPredicate(id);
        String result = "1";
        result = evaluatePredicateWithErrors(predicate, getAttributesMap(id));
        return result;
    }

    public String getPredicate(UUID id) {
        if(nodePredicates.containsKey(id)) {
            return nodePredicates.get(id);
        }
        if(id.equals(storage.getRootNode())) {
            return srcTree.getNode(id).getPredicate();
        }

        String full = "";
        if(getParent(id)!=null) {
            full = getPredicate(getParent(id));
        }
        String newPart = srcTree.getNode(id).getPredicate();

        return concatPredicate(full, newPart);
    }
    
    private String concatPredicate(String full, String newPart) {
        if(full.length()==0) return newPart;
        if(newPart.length()==0)  return full;
            
        if(full.equals(newPart)) return full;
        return "("+full+") && ("+newPart+")";
    }
    
    private Map<String, Object> getAttributesMap(UUID id) {
        if(nodeAttribute.containsKey(id)) {
            return nodeAttribute.get(id);
        }
        Map<String, Object> result = new HashMap<String, Object>(); 
        if(getParent(id) != null && !id.equals(storage.getRootNode())) {
            result.putAll(getAttributesMap(getParent(id)));
        } else {
            result.putAll(initalAttribute);
        }

        for(String key : srcTree.getNode(id).getAttributeKeys()) {
            result.put(key, srcTree.getNode(id).getAttributeValue(key).toString());
        }
        return result;
    }

    private boolean evaluatePredicate(String predicate, Map<String, Object> map) {
    	String s = evaluatePredicateWithErrors(predicate, map);
    	return s.charAt(0)=='1';
    }
    private String evaluatePredicateWithErrors(String predicate, Map<String, Object> map) {
        if(predicate==null || predicate.length()==0) return "1";
        String error = "";
        Boolean contextResult=true;
        try {
            Context context = Context.enter();
            WrapFactory wrapFactory = new WrapFactory();
            context.setWrapFactory(wrapFactory);
            Scriptable scope = createScope(context, map);
            Object o = context.evaluateString( scope, predicate, "", 1, null );
            if(o instanceof Boolean)
            	contextResult = (Boolean)o;
            else{
            	if(o instanceof NativeJavaObject){
            		o = ((NativeJavaObject)o).unwrap();
            		if(o instanceof Boolean)
            			contextResult = (Boolean)o;
            		else
            			error+="'"+predicate+"' returns not boolean value";
            	}else
            		error+="'"+predicate+"' returns undefined value";
            }
        } catch(Exception e){
        	error+="'"+predicate+"' causes an error: "+e.getMessage();
        } finally {
            Context.exit();
        }
        return (contextResult?"1":"0")+error;
    }
    
    private Scriptable createScope(Context context, Map<String, Object> attributes) {
        Scriptable topScope = ScriptRuntime.initStandardObjects(context, null, true);
        Scriptable scope = new NativeObject();
        scope.setParentScope( null );
        scope.setPrototype( topScope );

        for(String attrName: attributes.keySet() ) {
            String value = attributes.get(attrName).toString();
            Object obj = context.getWrapFactory().wrap( context, scope, value, value.getClass() );
            scope.put( attrName, scope, obj );
        }

        return scope;
    }

    @Override
    public void createNode(UUID parentUUId, UUID uuid, String id,
            Collection<Attribute> attrs) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void remove(UUID node) {
        throw new UnsupportedOperationException();
    }

    @Override
    public String getNodeId(UUID uuid) {
        return storage.getNodeId(uuid);
    }

    @Override
    public void changeNodeId(UUID node, String id) {
        throw new UnsupportedOperationException();
    }

    @Override
    public Collection<Attribute> getAttributes(UUID uuid) {
    	Map <String, Attribute> map = new HashMap<String, Attribute>();
    	if(changedAttribute.containsKey(uuid))
			for(Attribute a:changedAttribute.get(uuid).values())
				if(!map.containsKey(a.getKey()))
					map.put(a.getKey(), a);
    	for(Attribute a:storage.getAttributes(uuid))
    		if(!map.containsKey(a.getKey()))
    			map.put(a.getKey(), a);
    	return map.values();
    }

    @Override
    public void saveAttributes(UUID node, Collection<Attribute> attributes) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void moveNodeTo(UUID node, UUID parent) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean canLoad(UUID uuid) {
        return storage.canLoad(uuid);
    }

    @Override
    public String getNodeResourcePath(UUID uuId) {
        return storage.getNodeResourcePath(uuId);
    }

    @Override
    public InputStream getResourceContents(UUID uuid, String resName) {
        return storage.getResourceContents(uuid, resName);
    }

    @Override
    public void setResourceContents(UUID uuid, String resName, InputStream in, Object data) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean isResourceExists(UUID uuid, String resName) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void deleteResource(UUID uuid, String resName) {
        // TODO: instancer resource support  
        throw new UnsupportedOperationException();
    }

    @Override
    public String[] listResources(UUID uuid) {
        // TODO: instancer resource support
        throw new UnsupportedOperationException();
    }

    @Override
    public void createStorage(String name) {
        // TODO Auto-generated method stub
    }

    @Override
    public void deleteStorage() {
        // TODO Implement me!
        throw new UnsupportedOperationException("not implemented");
    }

    @Override
    public String getName() {
        return ":"+srcTree.getName();
    }
}
