/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.core;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Scanner;
import java.util.UUID;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IProjectDescription;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.IWorkspaceRunnable;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.jobs.ISchedulingRule;
import org.json.JSONException;

import com.unitesk.requality.core.Attribute;
import com.unitesk.requality.core.AttributeType;
import com.unitesk.requality.core.NodeMapper;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.AbstractTreeStorage;
import com.unitesk.requality.eclipse.natures.ReqCatalogue;
import com.unitesk.requality.nodetypes.ErrorNode;

public class ResourceStorage extends AbstractTreeStorage {
    public static final String RESOURCE_POSTFIX = "_resources";
    private static final String ROOT_NODE_FILENAME = "root.json";
    private IProject project;
    private UUID rootNodeUUID = null;

    private DeltaProcessor deltaProcessor = new DeltaProcessor(this);

    public IResourceChangeListener getResourceChangeListener() {
        return deltaProcessor;
    }

    public DeltaProcessor getDeltaProcessor() {
        return deltaProcessor;
    }

    public ResourceStorage(IProject project) {
        this.project = project;
        TreesTracker.getTracker().assignTreeDB(project, this);
    }

    public IFolder getResourceFolder(UUID uuid) {
        String id = getNodeId(uuid);
        IFolder resFolder = getResource(uuid).getParent().getFolder(
                new Path(id + RESOURCE_POSTFIX));
        return resFolder;
    }

    @Override
    public synchronized UUID getRootNode() {
        if (rootNodeUUID != null) {
            return rootNodeUUID;
        }
        IFile file = project.getFile(ROOT_NODE_FILENAME);
        if (file.exists()) {
            if (!resourceMap.inCache(file)) {
                getUUId(file); //resourceMap.cachePut(file, uuid);
            }
            rootNodeUUID = resourceMap.cacheGet(file);
            return rootNodeUUID;
        }
        return null;
    }

    @Override
    public String getNodeId(UUID uuid) {
        if (uuid.equals(getRootNode())) {
            return "/";
        }
        IFile file = getResource(uuid);
        if (file == null) {
            return null; // node not found
        }
        String jext = file.getFileExtension();
        String fileName = file.getName();
        return fileName.substring(0, fileName.length() - (jext.length() + 1));
    }

    @Override
    public Collection<Attribute> getAttributes(UUID uuid) {
        IFile file = getResource(uuid);
        String nodeSource = readFile(file);
        try {
            return NodeMapper.getAttributes(nodeSource);
        } catch (JSONException e) {
            StringWriter stackTrace = new StringWriter();
            e.printStackTrace(new PrintWriter(stackTrace));
            ArrayList<Attribute> result = new ArrayList<Attribute>();
            result.add(new Attribute(AttributeType.STRING,
                    TreeNode.ATTR_DESCRIPTION, stackTrace.toString()));
            result.add(new Attribute(AttributeType.STRING, TreeNode.ATTR_TYPE,
                    ErrorNode.TYPE_NAME));
            return result;
        }
    }

    @Override
    public UUID getParent(UUID uuid) {
        if (getResource(uuid) == null) {
            return null; // node not found
        }
        return getParent(getResource(uuid));
    }

    public UUID getParent(IFile file) {
        if (file.getParent().getType() == IResource.PROJECT) {
            return null;
        }
        String folderName = file.getParent().getName();
        String parentId = folderName.substring(0, folderName.length());

        IFile parentResource;
        IContainer parentFolder = file.getParent().getParent();
        parentResource = parentFolder.getFile(new Path(parentId + ".json"));
        if (!parentResource.exists()) {
            if (resourceMap.inCache(parentResource)) {
                return resourceMap.cacheGet(parentResource);
            }
            return null;
        }
        return getUUId(parentResource);
    }

    public UUID getUUId(IFile file) {
        if (!resourceMap.inCache(file)) {
            try {
                UUID uuid = NodeMapper.getUUId(readFile(file));
                if (uuid != null) {
                    resourceMap.cachePut(file, uuid);
                }
                return uuid;
            } catch (JSONException e) {

                // e.printStackTrace();
                // random uuid for error node
                return UUID.randomUUID();
            }
        } else {
            return resourceMap.cacheGet(file);
        }
    }

    @Override
    public synchronized Collection<UUID> getChildren(UUID parentUUId) {
        Collection<UUID> result = new HashSet<UUID>();
        IFile parentFile = getResource(parentUUId);
        if (parentFile == null) {
            return result; // node not found
        }

        String nodeId = getNodeId(parentUUId);
        if (nodeId.equals("/")) {
            nodeId = "root";
        }

        IFolder childrenFolder = parentFile.getParent().getFolder(
                Path.fromOSString(nodeId));

        if (!childrenFolder.exists()) {
            return result;
        }

        try {
            for (IResource res : childrenFolder.members()) {
                if (res.getType() == IResource.FILE
                        && res.getName().endsWith(".json")) {
                    UUID childUUID = getUUId((IFile) res);
                    result.add(childUUID);
                }
            }
        } catch (CoreException e) {
            e.printStackTrace();
        }
        return result;
    }

    @Override
    public boolean canLoad(UUID uuid) {
        return resourceMap.inCache(uuid);
    }

    private String readFile(IFile file) {
        try {
            if (!file.isSynchronized(IResource.DEPTH_ZERO)) {
                // TODO: add monitor
                file.refreshLocal(IResource.DEPTH_ZERO, null);
            }
            String fileExt = file.getFileExtension();
            if (fileExt == null || !fileExt.equals("json")) {
                return null;
            }
            StringBuilder jsonsb = new StringBuilder();
            Scanner s = new Scanner(file.getContents(), file.getCharset());
            while (s.hasNextLine()) {
                jsonsb.append(s.nextLine().trim());
            }
            s.close();
            return jsonsb.toString();
        } catch (CoreException e) {
            e.printStackTrace();
        }
        return null;
    }



    @Override
    public String getNodeResourcePath(UUID uuId) {
        return getResourceFolder(uuId).getLocation().toOSString();
    }

    public IProject getProject() {
        return project;
    }

    @Override
    public InputStream getResourceContents(UUID uuid, String resName) {
        try {
            // TODO why we lost sync here?
            IFolder resFolder = getResourceFolder(uuid);
            if (!resFolder.isSynchronized(IResource.DEPTH_INFINITE)) {
                resFolder.refreshLocal(IResource.DEPTH_INFINITE, null);
            }
            return resFolder.getFile(resName).getContents();
        } catch (CoreException e) {
            e.printStackTrace();
        }
        return null;
    }

    @Override
    public boolean isResourceExists(UUID uuid, String resName) {
        return getResourceFolder(uuid).getFile(resName).exists();
    }

    private void recursiveResourceList(IFolder folder, String path,
            ArrayList<String> result) throws CoreException {
        for (IResource res : folder.members()) {
            // use unix path separator
            String curPath = path + "/" + res.getName();
            if (res instanceof IFolder) {
                recursiveResourceList((IFolder) res, curPath, result);
            } else {
                result.add(curPath);
            }
        }
    }

    @Override
    public String[] listResources(UUID uuid) {
        ArrayList<String> result = new ArrayList<String>();
        try {
        	IFolder ifld = getResourceFolder(uuid);
        	if(ifld != null && ifld.exists())
        		recursiveResourceList(ifld, "", result);
        } catch (CoreException e) {
            e.printStackTrace();
        }
        return result.toArray(new String[0]);
    }

    @Override
    public String getName() {
        return project.getName();
    }

    // //////////////////////////////////////////////////////////////////////////
    //
    // Workspace operations
    //
    // //////////////////////////////////////////////////////////////////////////
    @Override
    public void createNode(final UUID parentUUId, final UUID uuid,
            final String id, final Collection<Attribute> attrs) {
        IWorkspaceRunnable createJob = new IWorkspaceRunnable() {
            @Override
            public void run(IProgressMonitor monitor) throws CoreException {
                try {
                    IFile file;
                    if (parentUUId == null) {
                        file = project.getFile(ROOT_NODE_FILENAME);
                    } else {
                        // get parent id
                        String parentId = getNodeId(parentUUId);
                        if (parentId.equals("/")) {
                            parentId = "root";
                        }
                        IFolder childFolder;
                        IFile parentFile = getResource(parentUUId);
                        if (parentFile.getParent().getType() == IResource.PROJECT) {
                            childFolder = ((IProject) parentFile
                                    .getParent()).getFolder(parentId);
                        } else {
                            childFolder = ((IFolder) parentFile
                                    .getParent()).getFolder(parentId);
                        }
                        if (!childFolder.exists()) {
                            childFolder.create(false, true, monitor);
                        }
                        file = childFolder.getFile(id + ".json");
                    }

                    String jsonString = NodeMapper.toJSON(uuid, attrs)
                            .toString(4);
                    ByteArrayInputStream in = new ByteArrayInputStream(
                            jsonString.getBytes("UTF-8"));

                    synchronized (ResourceStorage.this) {
                        resourceMap.cachePut(file, uuid);
                    }

                    file.create(in, false, monitor);
                    if (!file.getCharset().equals("UTF-8")) {
                        file.setCharset("UTF-8", monitor);
                    }
                    getDeltaTarget().created(uuid, true);
                } catch (UnsupportedEncodingException e) {
                    e.printStackTrace();
                } catch (JSONException e) {
                    e.printStackTrace();
                }
            }
        };

        ISchedulingRule rule = ResourcesPlugin.getWorkspace().getRoot();
        runWithoutDeltaProc(createJob, rule);
    }

    @Override
    public void changeNodeId(final UUID uuid, final String id) {
        IWorkspaceRunnable changeNodeId = new IWorkspaceRunnable() {
            @Override
            public void run(IProgressMonitor monitor) throws CoreException {
                String nodeId = getNodeId(uuid);
                IFile nodeFile = getResource(uuid);
                IPath newPath = nodeFile.getParent().getFullPath();
                newPath = newPath.addTrailingSeparator();
                IPath newFilePath = newPath.append(id + ".json");
                nodeFile.move(newFilePath, false, monitor);
                IFolder folder = project.getFolder(newPath.append(nodeId + RESOURCE_POSTFIX)
                        .makeRelativeTo(project.getFullPath()));
                if (folder.exists()) {
                    folder.move(newPath.append(id + RESOURCE_POSTFIX), false, monitor);
                }
                folder = project.getFolder(newPath.append(nodeId)
                        .makeRelativeTo(project.getFullPath()));
                if (folder.exists()) {
                    folder.move(newPath.append(id), true, monitor);
                }
                IFile newFile = project.getFile(newFilePath
                        .makeRelativeTo(project.getFullPath()));
                resourceMap.cacheReplace(nodeFile, newFile, uuid);
                resourceMap.updatePath(nodeFile.getParent().getLocation().append(nodeId),
                        newFile.getParent().getLocation().append(id));
                getDeltaTarget().changed(uuid, true);
            }


        };
        ISchedulingRule rule = ResourcesPlugin.getWorkspace().getRoot();
        runWithoutDeltaProc(changeNodeId, rule);
    }


    @Override
    public void remove(final UUID uuid) {
        // TODO: remove with new delta processor
        Collection<UUID> children = getChildren(uuid);
        for (UUID child : children) {
            remove(child);
        }
        final IFile res = getResource(uuid);
        IWorkspaceRunnable deleteJob = new IWorkspaceRunnable() {
            @Override
            public void run(IProgressMonitor monitor) throws CoreException {
                String id = getNodeId(uuid);
                IContainer parent = res.getParent();
                IFolder children = parent.getFolder(new Path(id));
                IFolder resources = parent.getFolder(new Path(id + RESOURCE_POSTFIX));
                if (resources.exists()) {
                    resources.delete(false, monitor);
                }
                if (children.exists()) {
                    children.delete(false, monitor);
                }
                res.delete(false, monitor);
                if (parent instanceof IFolder && parent.members().length == 0) {
                    parent.delete(false, monitor);
                }
                getDeltaTarget().deleted(uuid, true);
                resourceMap.cacheRemove(res);
            }
        };
        ISchedulingRule rule = ResourcesPlugin.getWorkspace().getRoot();
        runWithoutDeltaProc(deleteJob, rule);
    }

    @Override
    public void saveAttributes(final UUID uuid,
            final Collection<Attribute> attributes) {
        IWorkspaceRunnable saveAttributes = new IWorkspaceRunnable() {
            @Override
            public void run(IProgressMonitor monitor) {
                String jsonString;
                try {
                    jsonString = NodeMapper.toJSON(uuid, attributes)
                            .toString(4);
                    ByteArrayInputStream in = new ByteArrayInputStream(
                            jsonString.getBytes("UTF-8"));
                    getResource(uuid).setContents(in, true, true, monitor);
                    getDeltaTarget().changed(uuid, true);
                } catch (JSONException e) {
                    e.printStackTrace();
                } catch (CoreException e) {
                    e.printStackTrace();
                } catch (UnsupportedEncodingException e) {
                    e.printStackTrace();
                }
            }
        };
        ISchedulingRule rule = ResourcesPlugin.getWorkspace().getRoot();
        runWithoutDeltaProc(saveAttributes, rule);
    }

    @Override
    public void moveNodeTo(final UUID curNodeUUId, final UUID newParentUUId) {
        UUID oldParentUUID = getParent(curNodeUUId);
        String parentId = getNodeId(newParentUUId);
        if (parentId.equals("/")) {
            parentId = "root";
        }
        String nodeId = getNodeId(curNodeUUId);
        IFile toFile = getResource(newParentUUId).getParent()
                .getFolder(Path.fromOSString(parentId))
                .getFile(Path.fromOSString(nodeId + ".json"));
        IWorkspaceRunnable moveJob = new MoveNodeJob(this, toFile, curNodeUUId,
                oldParentUUID, newParentUUId);

        ISchedulingRule rule = ResourcesPlugin.getWorkspace().getRoot();
        runWithoutDeltaProc(moveJob, rule);
    }
    private void recCreate(IContainer rt, IProgressMonitor monitor) throws CoreException{
    	if(!rt.exists()){
    		recCreate(rt.getParent(), monitor);
    		if(rt instanceof IFolder)
    			((IFolder)rt).create(false, true, monitor);
    	}
    }
    @Override
    public void setResourceContents(final UUID uuid, final String resName,
            final InputStream in, Object data) {
        IWorkspaceRunnable setResourceContents = new IWorkspaceRunnable() {
            @Override
            public void run(IProgressMonitor monitor) throws CoreException {
                IFolder folder = getResourceFolder(uuid);
                if (!folder.exists()) {
                    folder.create(false, true, monitor);
                }
                IFile file = folder.getFile(resName);
                if (!file.getParent().exists())
                	recCreate(file.getParent(), monitor);
                if (!file.exists()) {
                    file.create(in, false, monitor);
                } else {
                    file.setContents(in, false, false, monitor);
                }
                getDeltaTarget().changed(uuid, true);
            }
        };

        ISchedulingRule rule = ResourcesPlugin.getWorkspace().getRoot();
        runWithoutDeltaProc(setResourceContents, rule);
    }

    @Override
    public void deleteResource(final UUID uuid, final String resName) {
        IWorkspaceRunnable deleteResource = new IWorkspaceRunnable() {
            @Override
            public void run(IProgressMonitor monitor) throws CoreException {
                IFolder res_folder = getResourceFolder(uuid);
                res_folder.getFile(resName).delete(false, monitor);
                if (res_folder.members().length == 0) {
                    res_folder.delete(false, monitor);
                }
                getDeltaTarget().changed(uuid, true);
            }
        };

        ISchedulingRule rule = ResourcesPlugin.getWorkspace().getRoot();
        runWithoutDeltaProc(deleteResource, rule);
    }

    @Override
    public void createStorage(String name) {
        IWorkspaceRunnable createStorage = new IWorkspaceRunnable() {
            @Override
            public void run(IProgressMonitor monitor) throws CoreException {
                project.create(monitor);
                project.open(monitor);
                IProjectDescription description = project.getDescription();
                String[] natures = description.getNatureIds();
                String[] newNatures = new String[natures.length + 1];
                System.arraycopy(natures, 0, newNatures, 0, natures.length);
                newNatures[natures.length] = ReqCatalogue.ID;
                description.setNatureIds(newNatures);
                project.setDescription(description, monitor);
            }
        };
        ISchedulingRule rule = ResourcesPlugin.getWorkspace().getRoot();
        runWithoutDeltaProc(createStorage, rule);
    }

    @Override
    public void deleteStorage() {
        IWorkspaceRunnable deleteStorage = new IWorkspaceRunnable() {
            @Override
            public void run(IProgressMonitor monitor) throws CoreException {
                project.delete(false, null);
            }
        };
        ISchedulingRule rule = ResourcesPlugin.getWorkspace().getRoot();
        runWithoutDeltaProc(deleteStorage, rule);
    }

    private void runWithoutDeltaProc(IWorkspaceRunnable job,
            ISchedulingRule rule) {
        try {
            // TODO: remove hack with delta processor activation
            deltaProcessor.deactivate();
            ResourcesPlugin.getWorkspace().run(job, rule,
                    IWorkspace.AVOID_UPDATE, null);
        } catch (CoreException e) {
            e.printStackTrace();
        } finally {
            deltaProcessor.activate();
        }
    }

    // //////////////////////////////////////////////////////////////////////////
    //
    // cache related methods
    //
    // //////////////////////////////////////////////////////////////////////////
    private ResourceMap resourceMap = new ResourceMap();
    public ResourceMap getResourceMap() {
        return resourceMap;
    }
    public static class ResourceMap {
        private Map<IFile, UUID> resourceToNode = new HashMap<IFile, UUID>();

        public synchronized void cacheRemove(IFile iFile) {
            resourceToNode.remove(iFile);
        }

        public synchronized boolean inCache(IResource resource) {
            return resourceToNode.containsKey(resource);
        }

        public synchronized boolean inCache(UUID uuid) {
            return resourceToNode.containsValue(uuid);
        }

        public synchronized void cachePut(IFile iFile, UUID uuid) {
            resourceToNode.put(iFile, uuid);
        }

        public synchronized UUID cacheGet(IFile iFile) {
            return resourceToNode.get(iFile);
        }

        public synchronized void cacheReplace(IFile oldFile, IFile newFile,
                UUID uuid) {
            resourceToNode.remove(oldFile);
            resourceToNode.put(newFile, uuid);
        }

        public synchronized void updatePath(IPath oldParentLocation,
                IPath newParentLocation) {
            ArrayList<IResource> resources = new ArrayList<IResource>(
                    resourceToNode.keySet());
            for (IResource res : resources) {
                IPath curLocation = res.getLocation();
                if (oldParentLocation.isPrefixOf(curLocation)) {
                    IPath newLocation = curLocation
                            .removeFirstSegments(oldParentLocation
                                    .segmentCount());
                    newLocation = newParentLocation.append(newLocation);
                    IFile newFile = ResourcesPlugin.getWorkspace()
                            .getRoot().getFileForLocation(newLocation);
                    UUID uuid = resourceToNode.remove(res);
                    resourceToNode.put(newFile, uuid);
                }
            }
        }

        public synchronized IFile getResource(UUID uuid) {
            for (IFile r : resourceToNode.keySet()) {
                if (resourceToNode.get(r).equals(uuid))
                    return r;
            }
            return null;
        }

    }

    public synchronized IFile getResource(UUID uuid) {
        return resourceMap.getResource(uuid);
    }
}
